/*
 * This File is part of the iowj-project   
 * $Id: AbstractIowDevcie.java,v 1.1.1.1 2006/01/15 16:47:26 Thomas Wagner Exp $
 * Copyright (C)2003-2006 by Thomas Wagner
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
package de.wagner_ibw.iow;

import java.util.NoSuchElementException;
import java.util.Vector;

import sun.misc.Queue;

import com.codemercs.iow.IowKit;

import de.wagner_ibw.iow.i2c.Monitor;

/**
 * This class represents the iow 40 device.
 * @author Thomas Wagner
 */
public abstract class AbstractIowDevcie implements Runnable{

	/** 
	 * A constant for iow 40 device id.
	 * (decimal 5376)
	 */
	public static final long IOW40ID = 0x1500;
	
	/**
	 * A constant for iow 24 device id.
	 * (decimal 5377)
	 */	
	public static final long IOW24ID = 0x1501;
	
	/**
	 * A constant for iow 40 device name. 
	 */	
	public static final String IOW40NAME = "IOW40";
	
	/**
	 * A constant for iow 24 device name.
	 */	
	public static final String IOW24NAME = "IOW24";

	/**
	 * Device handle of this iow device.
	 */
	protected long handle;
	
	/**
	 * Device identifier of this iow device.
	 */
	protected int id;
	
	/**
	 * Revision identifier of this iow device.
	 */
	protected int rev;
	
	/**
	 * Serial number of this iow device.
	 */
	protected String serial;

	/**
	 * The ports.
	 */	
	protected IowPort[] ports = {new IowPort(0,"IIIIIIII"),
					   new IowPort(1,"IIIIIIII"),
					   new IowPort(2,"IIIIIIII"),
					   new IowPort(3,"IIIIIIII")};

	
	/**
	 * Queue for receiving the current pin status report
	 * @since 0.9.3
	 */
	protected Queue currentPinStatus;

	/**
	 * Queue for receiving the current pin status report
	 * @since 0.9.4
	 */
	protected Monitor monitor;

	/**
	 * Flag for all special mode functions. Each bit stands for one enabled or
	 * disbaled special mode function. See SpecialModeFunction constants.
	 * @see de.wagner_ibw.iow.SpecialModeFuntion
	 */
	protected int specialMode = 0;

	/**
	 * Vector of special mode implementations.
	 */
	protected Vector smfImplementations = new Vector();		
	
	protected boolean autonomous = false;    
	
	protected int portCount = 0;
	
	/**
	 * Contructor.
	 *
	 */
	public AbstractIowDevcie(){
		currentPinStatus = new Queue();
		monitor = new Monitor();
	}
	
	/**
	 * 
	 * @param impl
	 */
	public synchronized void addSpecialModeFunctionImpl(SpecialModeFunction impl)
		throws UnsupportedOperationException, IllegalArgumentException {
		
		String error = impl.checkCompatibility(id, rev, specialMode);
		if(error != null)
			throw new UnsupportedOperationException(AbstractIowDevcie.IOW24NAME + ", rev " + Integer.toHexString(rev) + " does not support " + impl.getName() + " at this time: " + error + "!");
		
		if (!smfImplementations.contains(impl)) {
			//set related special mode bit
			specialMode = specialMode | impl.getSpecialModeFuncionId();
			//enabel special mode function
			long ret = writeReport(1, impl.getEnableReport()); //check ret!
			//set special bits in bitmask
			for (int i = 0; i < impl.getIowSpecialBits(id).length; i++) {
				ports[i].setSpec(ports[i].getSpec() | impl.getIowSpecialBits(id)[i]);
			}
			//set iow reference
			impl.setIowDevice(this);
			//add implementation to list
			smfImplementations.add(impl);
		} else {
			throw new IllegalArgumentException(impl.getName() + " was alreay added!");
		}
	}

	/**
	 * @param impl
	 */
	public synchronized void removeSpecialModeFunctionImpl(SpecialModeFunction impl) {
		if (smfImplementations.contains(impl)) {
			//reset related special mode bit
			specialMode = specialMode & (255 - impl.getSpecialModeFuncionId());
			//disable special mode function		
			long ret = writeReport(1, impl.getDisableReport()); //check ret!
			//reset special bits in bitmask
			for (int i = 0; i < impl.getIowSpecialBits(id).length; i++) {
				ports[i].setSpec(ports[i].getSpec() & (255 - impl.getIowSpecialBits(id)[i]));
			};
			//delete iow reference
			impl.setIowDevice(null);
			//remove implementation from list
			smfImplementations.remove(impl);
		}
	}	
	
	/**
	 * Returns the desired special mode function implementation.
	 * @param smfName Name of the desired function. 
	 * @return Instance of the desired special mode function implementation.
	 * @throws NoSuchElementException if the desired function was not already added.
	 * @since 0.9.4
	 */
	public SpecialModeFunction getSpecialModeFunctionImpl(String smfName)
		throws NoSuchElementException {
		
		for (int i = 0, length = smfImplementations.size(); i < length; i++) {
			SpecialModeFunction smf = (SpecialModeFunction) smfImplementations.get(i);
			if (smf.getName().equals(smfName))
				return smf;
		}
		throw new NoSuchElementException("Special mode function "+ smfName + " not found!");	
	}
	
	/**
	 * Implementation for IowInterface method getHandle.
	 * @return Returns the handle of this iow device.
	 */	
	public long getHandle(){
		return handle;
	}
	
	/**
	 * Implementation for IowInterface method getSerial.
	 * @return The serial number of this iow device.
	 */
	public String getSerial(){
		return serial;
	}
	
	/**
	 * Implementation for IowInterface method getId.
	 * @return Returns the device id of this device.
	 */
	public int getId(){
		return id;
	}
	
	/**
	 * Implementation for IowInterface method getRev.
	 * @return Returns the revision id of this device.
	 * @since 0.9.3
	 */
	public int getRev(){
		return rev;
	}

	/**
	 * Close this iow device
	 * (Used by subclasses Iow40 and Iow24 only.)
	 */
	public void close(){
		if(handle != 0) {
			IowKit.closeDevice(handle);
			handle = 0;
		}
		else {
			System.out.println("There is no open device!");
		}
	}

	/**
	 * Set timeout for read operations.
	 * @param timeout Timeout value in ms.
	 */
	public void setTimeout(long timeout) {
		IowKit.setTimeout(handle,timeout);
	}
	
	/**
	 * Set timeout for write operations.
	 * @param timeout Timeout value in ms.
	 * @since 0.9.3
	 */
	public void setWriteTimeout(long timeout) {
		IowKit.setWriteTimeout(handle,timeout);
		
	}

	/**
	 * Implementation for IowInterface method readReport.
	 * @param pipe
	 * @return Array of int.
	 */
	public int[] readReport(int pipe){
		int rbuf[] = IowKit.read(handle,pipe,8);
		
		//int len = rbuf.length;
		for (int i = 0, len = rbuf.length ; i < len; i++) {
			rbuf[i] = rbuf[i] & 0xff; 
		}
		return rbuf;
	}
		
	/**
	 * Implementation for IowInterface method writeReport.
	 * @param pipe
	 * @param wbuf Write buffer.
	 * @return Count of written bytes.
	 */	
	public long writeReport(int pipe, int[] wbuf){
		return IowKit.write(handle,pipe,wbuf);
	}


	/**
	 * @since 0.9.3 with long return value
	 */
	public abstract long readIOPorts();
	
	/**
	 * @since 0.9.3 with long return value
	 */
	public abstract long readIOPortsImmediate();
	
	/**
	 * Gets the current pin status.
	 * @return Count of written bytes to iow device.
	 * @since 0.9.3
	 */
	public abstract long scanPorts() throws UnsupportedOperationException;
	
	/**
	 * Gets the current pin status.
	 * @return Array of int.
	 * @since 0.9.3
	 */
	protected int[] getCurrentPinStatus() throws UnsupportedOperationException {
		int wbuf[] = {0xff,0,0,0,0,0,0,0};
		
		if(rev < 0x1011) {
			throw new UnsupportedOperationException("Release " + Integer.toHexString(rev) + " does not support the current pin status report");
		}
										
		long ret = writeReport(1,wbuf);
		if(ret == 8){
			int rbuf[] = {0,0,0,0};
			try {
				int[] resbuf = (int[])currentPinStatus.dequeue();
				rbuf[0] = resbuf[1];
				rbuf[1] = resbuf[2];
				rbuf[2] = resbuf[3];
				rbuf[3] = resbuf[4];
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
			return rbuf;
		}
		else {
			//*deb*/System.out.println("getCurrentPinStatus() writeReport returns != 8");
			return new int[0];
		}
	}
		
	public abstract void writeIOPorts();	
	
	public void run(){
		while (autonomous) {
			if (Thread.interrupted()) {
				break;
			}
			//*deb*/System.out.println("Thread: readIOPorts()...");
			readIOPorts();
			//*deb*/System.out.println("Thread: readIOPorts() done");
		}	
	}
	
	/**
	 * Sets the autonomous mode either on or off.
	 * Autonomous mode means, that class uses readIOPorts() in a loop an in a thread.
	 * @param status true switches the mode on, false switches the mod off.
	 */
	public void autonomous(boolean status){
		Thread t;
		if(status){
			autonomous = true;	
			t = new Thread(this);
			t.start();
		}
		else {
			autonomous = false;
		}
	}
	
	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.IowInterface#getName()
	 */
	public abstract String getName() ;
		
	public IowPort getPort(int port) throws IllegalArgumentException {
		checkPort(port);
		return ports[port];
	}		

 	/**
	 * Set the given bit from the given port.
	 * @param port Port number (0...3).
	 * @param bit Bit number (0...7).
	 */
	public void setBit(int port, int bit)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].setBit(bit);
	}
	
	/**
	 * Clear the given bit froma the given port.
	 * @param port Port number (0...3).
	 * @param bit Bit number (0...7).
	 */
	public void clearBit(int port, int bit)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].clearBit(bit);
	}

	/**
	 * Set the pattern to the given port.
	 * @param port Port number (0...3).
	 * @param pattern
	 */
	public void setPort(int port, int in)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].setData(in);
	}
	
	/**
	 * Set the direction mask to the given port.
	 * @param port Port number (0...3 or 0...1).
	 * @param direction 0-output, 1-input
	 * @since 0.9.3
	 */
	public void setDirection(int port, int direction)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].setDirection(direction);
	}
	
	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.IowInterface#getMonitor()
	 */
	public Monitor getMonitor() {
		return monitor;
	}

	protected void checkPort(int port) throws IllegalArgumentException {
		if(port < 0 || port > portCount )
			throw new IllegalArgumentException("Invalid port number (" + port + ")");
	}

	/** 
	 * Returns a hash code value for this Iow40 object. It is derived from the serial number.
	 * @return a hash code value for this object.
	 * @since 0.9.4
	 */
	public int hashCode() {
		int result = 17;
		result = 37 * result + id;
		result = 37 * result + serial.hashCode();
		return result;
	}

	/**
	 * Returns a <code>String</code> object representing this Iow40 object.
	 * It contains information about device handle, id , revision,
	 * serial number and the status of port 0 ... port 3.  
	 */
	public abstract String toString();
	
	/**
	 * Indicates whether some other Iow40 object is "equal to" this one.
	 * Criteria is the serial number.
	 * @param obj The reference object with which to compare.
	 * @return  True if this object is the same as the obj argument; false otherwise.
	 * @since 0.9.4
	 */
	public abstract boolean equals(Object o);

	/**
	 * Read thread for special mode functions.
	 * @author Thomas Wagner
	 */
	public class SpecialModeReadThread extends Thread {
			
		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			do {
				if (Thread.interrupted()) {
					break;
				}
				int[] rbuf = readReport(1);
				if( rbuf.length > 0) {
					//*deb*/System.out.println("Iow40.ReadThread.readReport(1): " + Integer.toHexString(rbuf[0]));
										
					for (int i = 0; i < smfImplementations.size(); i++) {
						SpecialModeFunction smf = (SpecialModeFunction) smfImplementations.get(i);
						if (smf.matchReportId(rbuf[0])) {
							smf.reportReceived(rbuf);
							break;
						}
					}
					if (rbuf[0] == 0xff) { //current pin status report
						//*deb*/System.out.println("current pin status report received !");
						currentPinStatus.enqueue(rbuf);
					}
				}
			} while (true);
		}
	}
}
