/*
 * This File is part of the iowj-project   
 * $Id: Iow40.java,v 1.1.1.1 2006/01/15 16:47:26 Thomas Wagner Exp $
 * Copyright (C)2004-2006 by Thomas Wagner
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
package de.wagner_ibw.iow;

import com.codemercs.iow.IowKit;

/**
 * This class represents the iow 40 device.
 * @author Thomas Wagner
 */
public class Iow40 extends AbstractIowDevcie {
		
	public Iow40(long handle, int id, String serial, int rev){
		portCount = 4;
		this.handle = handle;
		this.id = id;
		this.serial = serial;
		this.rev = rev;
		new SpecialModeReadThread().start();
	}

	/**
	 * @since 0.9.3 with long return value
	 */
	public long readIOPorts(){
		//*deb*/System.out.println("Iow40.readIOPorts()...");
		long ret = 0;
		int[] rbuf = IowKit.read(handle,0,5);
		for (int i = 1; i < rbuf.length; i++){
			int val = rbuf[i] & 0xff;
			ports[i-1].setDataFromRead(val);
			long newval = (long) val << ((i - 1) * 8);
			ret = ret + newval;
		}	
		return ret;
	}
	
	/**
	 * @since 0.9.3 with long return value
	 */
	public long readIOPortsImmediate(){
		int[] ri = IowKit.readImmediate(handle);
		long ret = 0;
		if(ri.length > 0) {
			//*deb*/System.out.println("Iow40.read immediate()->"+ ri[0]+","+ri[1]+","+ri[2]+","+ri[3]);
			for (int i = 0; i < ri.length; i++){
				ports[i].setDataFromRead(ri[i]);
				long newval = (long) ri[i] << (i * 8);
				ret = ret + newval;
			}
		}
		else {
			//*deb*/System.out.println("Iow40.read immediate()-> no result!");
		}
		return ret;
	}
	
	/**
	 * Gets the current pin status.
	 * @return Count of written bytes to iow device.
	 * @since 0.9.3
	 */
	public long scanPorts() throws UnsupportedOperationException {
		long ret = 0;
	
		int rbuf[] = getCurrentPinStatus();
		
		if(rbuf.length > 0) {
			for (int i = 0; i < rbuf.length; i++){
				int val = rbuf[i] & 0xff;
				ports[i].setDataFromRead(val);
				long newval = (long) val << (i * 8);
				ret = ret + newval;
			}	
		}	
		return ret;	
	}
	
	public void writeIOPorts() {
		if(ports[0].isChangeFlagSet() ||
		   ports[1].isChangeFlagSet() ||
		   ports[2].isChangeFlagSet() ||
		   ports[3].isChangeFlagSet()) {
			int[] wbuf = {0,0,0,0,0};
			wbuf[1] = ports[0].getDataToWrite();
			wbuf[2] = ports[1].getDataToWrite();
			wbuf[3] = ports[2].getDataToWrite();
			wbuf[4] = ports[3].getDataToWrite();
			long ret = IowKit.write(handle,0,wbuf);
			//*deb*/System.out.println("[Iow40.writeIOPorts() ret=" +ret);
			if(ret == 5) {
				ports[0].clearChangeFlag();
				ports[1].clearChangeFlag();
				ports[2].clearChangeFlag();
				ports[3].clearChangeFlag();
			}
		}
		else {
			//*deb*/System.out.println("No writeIOPorts() needed!");
		}
	}
	
	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.AbstractIowDevcie#getName()
	 */
	public String getName() {
		return IOW40NAME;
	}
	
	/**
	 * Returns a <code>String</code> object representing this Iow40 object.
	 * It contains information about device handle, id , revision,
	 * serial number and the status of port 0 ... port 3.  
	 */
	public String toString(){
		boolean i2c = false;
		boolean spi = false;
				
		StringBuffer sb = new StringBuffer();
		sb.append(getName());
		sb.append(":Handle[");
		sb.append(getHandle());
		sb.append("],Id[");
		sb.append(getId());
		sb.append("],Rev[");
		sb.append(Integer.toHexString(getRev()));
		sb.append("],Serial[");
		sb.append(getSerial());
		sb.append("],SpecialMode: ");
		if ((specialMode & SpecialModeFunction.SMF_LCD_ID) == SpecialModeFunction.SMF_LCD_ID)
			sb.append("LCD ");
		if ((specialMode & SpecialModeFunction.SMF_CPS_ID) == SpecialModeFunction.SMF_CPS_ID)
			sb.append("CPS ");
		if ((specialMode & SpecialModeFunction.SMF_SMX_ID) == SpecialModeFunction.SMF_SMX_ID)
			sb.append("SMX ");
		if ((specialMode & SpecialModeFunction.SMF_I2C_ID) == SpecialModeFunction.SMF_I2C_ID) {
			sb.append("I2C");
			i2c = true;		
		}
		sb.append("\n\t");
		sb.append(ports[0].toString());
		sb.append("\n\t");
		sb.append(ports[1].toString());
		sb.append("\n\t");
		sb.append(ports[2].toString());
		sb.append("\n\t");
		sb.append(ports[3].toString());
		if(i2c) {
			sb.append("\n\t");
			for (int i = 0; i < smfImplementations.size(); i++) {
				SpecialModeFunction smf = (SpecialModeFunction) smfImplementations.get(i);
				if (smf.getSpecialModeFuncionId() == SpecialModeFunction.SMF_I2C_ID) {
					sb.append(smf.toString());
					break;
				}
			}
		}	
		sb.append("\n");
		return sb.toString();
	}

	/**
	 * Indicates whether some other Iow40 object is "equal to" this one.
	 * Criteria is the serial number.
	 * @param obj The reference object with which to compare.
	 * @return  True if this object is the same as the obj argument; false otherwise.
	 * @since 0.9.4
	 */
	public boolean equals(Object o) {
		if(!(o instanceof Iow40))
			return false;
		
		Iow40 dev = (Iow40)o;
		return (this.serial.equals(dev.serial)) && (this.id == dev.id);
	}

}
