/*
 * This File is part of the iowj-project   
 * $Id: IowFactory.java,v 1.1.1.1 2006/01/15 16:47:26 Thomas Wagner Exp $
 * Copyright (C)2004 by Thomas Wagner
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.wagner_ibw.iow;

import java.io.ObjectStreamException;
import java.util.NoSuchElementException;
import java.util.Vector;

import com.codemercs.iow.IowKit;

/**
 * This is the base class of all the iow devices, based on the
 * com.codemercs.iow.IowKit JNI wrapper class to the iowkit library 
 * (iowki.dll for Windows, libiowkit.so for Linux).
 * @author Thomas Wagner
 */
public class IowFactory {
	
	/**
	 * Count of available iow devices.
	 */
	private static long devCount;
	
	/**
	 * Vector of all scanned iow devices.
	 */
	private static Vector iowDevices;
	
	public final static IowFactory INSTANCE = new IowFactory();

	private IowFactory() {
		iowDevices = new Vector();
		openAllDevices();
	}

	public static IowFactory getInstance() {
		return INSTANCE;
	}
		
	/**
	 * Ensure Singleton class
	 */
	 private Object readResolve() throws ObjectStreamException {
	   return INSTANCE;
	 }
			
	/**
	 * Open all connected iow devices.
	 * @since 0.9.3
	 */
	private long openAllDevices() {
		IowKit.openDevice();
		//scan for iow devices
		devCount    = IowKit.getNumDevs();
		for (int i = 0; i < devCount; i++) {
			long handle = IowKit.getDeviceHandle(i+1); //Zhlung beginnt ab 1
			int id = (int) IowKit.getProductId(handle);
			String serial = IowKit.getSerialNumber(handle);
			int rev = (int) IowKit.getRevision(handle);
			if(id == AbstractIowDevcie.IOW40ID){
				Iow40 dev40 = new Iow40(handle, id, serial, rev);
				iowDevices.add(dev40);
			}
			else if(id == AbstractIowDevcie.IOW24ID){
				Iow24 dev24 = new Iow24(handle, id, serial, rev);
				iowDevices.add(dev24);
			}
		}
		return devCount;
	}
	
	/**
	 * Returns the number of connected iow devices.
	 * @return Number of connected iow devices.
	 * @since 0.9.3
	 */
	public long getNumDevices() {
		return devCount;
	}
	
	/**
	 * Returns the first connected iow 40 device.
	 * @return Instance of an iow 40 device.
	 * @throws NoSuchElementException If no connected iow 40 device was found.
	 */
	public Iow40 getIow40Device() throws NoSuchElementException {
		Iow40 dev = null;
		boolean found = false;
		for(int i = 0; i < devCount; i++) {
			AbstractIowDevcie iowd = (AbstractIowDevcie) iowDevices.get(i);
			if( iowd.getId() == AbstractIowDevcie.IOW40ID) {
				dev = (Iow40) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException(AbstractIowDevcie.IOW40NAME);
		}
		return dev;
	}
		
	/**
	 * Returns the iow40 device according to the specified serial number.
	 * @param serial Serial number of the desired iow 40 device.
	 * @return @return Instance of the desired iow 40 device.
	 * @throws NoSuchElementException If no iow 40 device with desired serial number was found.
	 */
	public Iow40 getIow40Device(String serial) throws NoSuchElementException {
		Iow40 dev = null;
		boolean found = false;
		for(int i = 0; i < devCount; i++) {
			AbstractIowDevcie iowd = (AbstractIowDevcie) iowDevices.get(i);
			if( iowd.getId() == AbstractIowDevcie.IOW40ID && iowd.getSerial().equalsIgnoreCase(serial)) {
				dev = (Iow40) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException("No " + AbstractIowDevcie.IOW40NAME + " device, serial=" + serial +" found!");
		}
		return dev;
	}		
	
	/**
	 * Returns the first connected iow 24 device.
	 * @return Instance of an iow 24 device.
	 * @throws NoSuchElementException If no connected iow 24 device was found.
	 */
	public Iow24 getIow24Device() throws NoSuchElementException {
		Iow24 dev = null;
		boolean found = false;
		for(int i = 0; i < devCount; i++){
			AbstractIowDevcie iowd = (AbstractIowDevcie) iowDevices.get(i);
			if( iowd.getId() == AbstractIowDevcie.IOW24ID){
				dev = (Iow24) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException(AbstractIowDevcie.IOW24NAME);
		}
		return dev;
	}

	/**
	 * Returns the iow 24 device according to the specified serial number.
 	 * @param serial Serial number of the desired iow 24 device.
 	 * @return Instance of the desired iow 24 device.
 	 * @throws NoSuchElementException If no iow 24 device with desired serial number was found.
	 */
	public Iow24 getIow24Device(String serial) throws NoSuchElementException {
		Iow24 dev = null;
		boolean found = false;
		for(int i = 0; i < devCount; i++){
			AbstractIowDevcie iowd = (AbstractIowDevcie) iowDevices.get(i);
			if( iowd.getId() == AbstractIowDevcie.IOW24ID && iowd.getSerial().equalsIgnoreCase(serial)) {
				dev = (Iow24) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException("No " + AbstractIowDevcie.IOW24NAME + " device, serial=" + serial +" found!");
		}
		return dev;
	}	
		
	/**
	 * Close all the iow devices.
	 */
	public void closeAllDevices(){
		IowKit.closeDevice(0L);
	}

	/**
 	 * Returns a string representation of the object. It consists of the
	 * information about all connected iow devices.
 	 * @return a string consisting of the device handle, id, revision and serial number of
 	 * each connected iow device.  
 	 */
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append("Number of scanned devices: " + devCount + "\n");
		for(int i = 0; i < devCount; i++){
			AbstractIowDevcie iowd = (AbstractIowDevcie) iowDevices.get(i);
			sb.append("Device" + (i+1) + ": ");
			sb.append(iowd.getName());
			sb.append(",Handle[");
			sb.append(iowd.getHandle());
			sb.append("],Id[");
			sb.append(iowd.getId());
			sb.append("],Rev[");
			sb.append(Integer.toHexString(iowd.getRev()));
			sb.append("],Serial[");
			sb.append(iowd.getSerial());
			sb.append("]\n");
		}
		return sb.toString();
	}
	
	/**
	 * Closes all open devices and terminates the jvm.
	 * @since 0.9.3
	 */
	public void exit(int code) {
		closeAllDevices();
		System.exit(code);
	}
	
	/**
	 * Returns the version of the loaded iowkit library.
	 * @return Version of the iowkit.dll.
	 * @since 0.9.3
	 */
	public String getVersion(){
		return IowKit.version();
	}
}
